/* FILE: input.c                Copyright (c), Symmetric Research, 1998
 *
 * DSPASM file input, lexical analysis, and symbol tables.
 *
 *
 *
 */

#include "dspasm.h"







/* INPUT FILE AND BUFFER FUNCTIONS:
 *
 *
 *
 *
 *
 */

#define MAXBYTES        (unsigned)0xFFF0        /* input buffer size    */
#define CTRL_Z          0x1A                    /* ^Z end of file mark  */

FILE *handle;                                   /* input file handle    */

char
        *input_buffer,                          /* -> input buffer      */
        *input,                                 /* -> current position  */
        *eob;                                   /* -> end of buffer     */


typedef struct {                                /* token data type      */

        unsigned

                nline,                          /* source line number   */
                type;                           /* token type           */

        char
                *line,                          /* -> source code line  */
                *start,                         /* -> token start       */
                *end;                           /* -> token end         */

        } TOKEN;


TOKEN

      current_token, token;  /* current and look ahead tokens */






int allocate_input_buffer( void) {


        /* Allocate the input buffer. */

        input_buffer = malloc( MAXBYTES);

          return( input_buffer ? 1 : 0 );


}
int read_input_file( void) {

        unsigned nbytes;



        /* Open and read the input file, setting the token pointers. */

        handle = fopen( filename, "rb");


        if ( handle ) {


                /* Read from the input file. */

                nbytes = fread( input_buffer, 1, MAXBYTES, handle);

                    eob = &input_buffer[nbytes]; *eob = CTRL_Z;



                /* Get the first token. */

                current_token.line  = input_buffer;
                current_token.nline = 1;
                current_token.start = input_buffer;
                current_token.end   = input_buffer;

                        token.line  = input_buffer;
                        token.nline = 1;
                        token.start = input_buffer;
                        token.end   = input_buffer;


                    next();


                return( 1);

                }


        else

                return( 0); /* failure */



}
int more_input_file( void) {

        unsigned nbytes, n; char *p1, *p2;




        /* Copy the current token line to input buffer start. */

        p1 = input_buffer; p2 = current_token.line;


        while ( p2 <= eob )             /* copy CTRL_Z eof mark too */

                *p1++ = *p2++;



        /* Fixup all the input buffer pointers to their new positions. */

        nbytes = current_token.line - input_buffer;

                      input -= nbytes;
                        eob -= nbytes;

        current_token.line  -= nbytes;
        current_token.start -= nbytes;
        current_token.end   -= nbytes;

                token.line  -= nbytes;
                token.start -= nbytes;
                token.end   -= nbytes;

        


        /* Try to read some more of the input file. */

        n = fread( eob, 1, nbytes, handle);


        if ( n ) {

                /* Set the new end of file and mark it. */

                eob += n; *eob = CTRL_Z;


                return( 1); /* success */

                }


        else

                return( 0); /* failure, no more file to read */





}
int reset_input_file( void) {

        unsigned nbytes;



        /* Reset the input file for a second pass. */

#define SUCCESS         0

        if ( fseek( handle, 0, SEEK_SET) == SUCCESS ) {


                /* Read from the input file. */
                
                nbytes = fread( input_buffer, 1, MAXBYTES, handle);

                    eob = &input_buffer[nbytes]; *eob = CTRL_Z;



                /* Get the first token. */

                current_token.line  = input_buffer;
                current_token.nline = 1;
                current_token.start = input_buffer;
                current_token.end   = input_buffer;

                        token.line  = input_buffer;
                        token.nline = 1;
                        token.start = input_buffer;
                        token.end   = input_buffer;


                    next();


                return( 1);

                }


        else

                return( 0); /* failure */


}
int nextchar( void) {


        /* Move on to the next char, checking for end of buffer. */

        if ( input < eob ) {

                input++;

                    if ( input == eob ) more_input_file();

                }


        return( *input);


}









/* CHARACTER CLASSIFICATION AND TOKEN GATHERING:
 *
 *
 *
 *
 *
 */

#define ALPHA           0x01                    /* leads to IDENT       */
#define ARITH           0x02                    /* leads to OPERT       */
#define DIGIT           0x04                    /* leads to NUMBR       */


#define IDENT           0x10                    /* token types          */
#define OPERT           0x20
#define NUMBR           0x40
#define FLOAT           0x80


static unsigned char type[] = {

        /* 00 NUL */    0                               ,
        /* 01 SOH */    0                               ,
        /* 02 STX */    0                               ,
        /* 03 ETX */    0                               ,
        /* 04 EOT */    0                               ,
        /* 05 ENQ */    0                               ,
        /* 06 ACK */    0                               ,
        /* 07 BEL */    0                               ,
        /* 08 BS  */    0                               ,
        /* 09 HT  */    0                               ,
        /* 0A LF  */    0                               ,
        /* 0B VT  */    0                               ,
        /* 0C FF  */    0                               ,
        /* 0D CR  */    0                               ,
        /* 0E SO  */    0                               ,
        /* 0F SI  */    0                               ,
        /* 10 DLE */    0                               ,
        /* 11 DC1 */    0                               ,
        /* 12 DC2 */    0                               ,
        /* 13 DC3 */    0                               ,
        /* 14 DC4 */    0                               ,
        /* 15 NAK */    0                               ,
        /* 16 SYN */    0                               ,
        /* 17 ETB */    0                               ,
        /* 18 CAN */    0                               ,
        /* 19 EM  */    0                               ,
        /* 1A SUB */    0                               ,
        /* 1B ESC */    0                               ,
        /* 1C FS  */    0                               ,
        /* 1D GS  */    0                               ,
        /* 1E RS  */    0                               ,
        /* 1F US  */    0                               ,
        /* 20 SP  */    0                               ,
        /* 21 !   */    ARITH   |OPERT                  ,
        /* 22 "   */    0                               ,
        /* 23 #   */    ARITH   |OPERT                  ,
        /* 24 $   */    0                               ,
        /* 25 %   */    0                               ,
        /* 26 &   */    ARITH   |OPERT                  ,
        /* 27 '   */    0                               ,
        /* 28 (   */    0                               ,
        /* 29 )   */    0                               ,
        /* 2A *   */    ARITH                           ,
        /* 2B +   */    ARITH   |OPERT          |FLOAT  ,
        /* 2C ,   */    0                               ,
        /* 2D -   */    ARITH   |OPERT          |FLOAT  ,
        /* 2E .   */    ARITH                           ,
        /* 2F /   */    ARITH   |OPERT                  ,
        /* 30 0   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 31 1   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 32 2   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 33 3   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 34 4   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 35 5   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 36 6   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 37 7   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 38 8   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 39 9   */    DIGIT   |NUMBR  |IDENT  |FLOAT  ,
        /* 3A :   */    0                               ,
        /* 3B ;   */    0                               ,
        /* 3C <   */    ARITH   |OPERT                  ,
        /* 3D =   */    ARITH   |OPERT                  ,
        /* 3E >   */    ARITH   |OPERT                  ,
        /* 3F ?   */    0                               ,
        /* 40 @   */    0                               ,
        /* 41 A   */    ALPHA   |IDENT  |NUMBR          ,
        /* 42 B   */    ALPHA   |IDENT  |NUMBR          ,
        /* 43 C   */    ALPHA   |IDENT  |NUMBR          ,
        /* 44 D   */    ALPHA   |IDENT  |NUMBR          ,
        /* 45 E   */    ALPHA   |IDENT  |NUMBR  |FLOAT  ,
        /* 46 F   */    ALPHA   |IDENT  |NUMBR          ,
        /* 47 G   */    ALPHA   |IDENT  |NUMBR          ,
        /* 48 H   */    ALPHA   |IDENT  |NUMBR          ,
        /* 49 I   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4A J   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4B K   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4C L   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4D M   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4E N   */    ALPHA   |IDENT  |NUMBR          ,
        /* 4F O   */    ALPHA   |IDENT  |NUMBR          ,
        /* 50 P   */    ALPHA   |IDENT  |NUMBR          ,
        /* 51 Q   */    ALPHA   |IDENT  |NUMBR          ,
        /* 52 R   */    ALPHA   |IDENT  |NUMBR          ,
        /* 53 S   */    ALPHA   |IDENT  |NUMBR          ,
        /* 54 T   */    ALPHA   |IDENT  |NUMBR          ,
        /* 55 U   */    ALPHA   |IDENT  |NUMBR          ,
        /* 56 V   */    ALPHA   |IDENT  |NUMBR          ,
        /* 57 W   */    ALPHA   |IDENT  |NUMBR          ,
        /* 58 X   */    ALPHA   |IDENT  |NUMBR          ,
        /* 59 Y   */    ALPHA   |IDENT  |NUMBR          ,
        /* 5A Z   */    ALPHA   |IDENT  |NUMBR          ,
        /* 5B [   */    0                               ,
        /* 5C \   */    0                               ,
        /* 5D ]   */    0                               ,
        /* 5E ^   */    ARITH   |OPERT                  ,
        /* 5F _   */    ALPHA   |IDENT  |NUMBR          ,
        /* 60 `   */    0                               ,
        /* 61 a   */    ALPHA   |IDENT  |NUMBR          ,
        /* 62 b   */    ALPHA   |IDENT  |NUMBR          ,
        /* 63 c   */    ALPHA   |IDENT  |NUMBR          ,
        /* 64 d   */    ALPHA   |IDENT  |NUMBR          ,
        /* 65 e   */    ALPHA   |IDENT  |NUMBR  |FLOAT  ,
        /* 66 f   */    ALPHA   |IDENT  |NUMBR          ,
        /* 67 g   */    ALPHA   |IDENT  |NUMBR          ,
        /* 68 h   */    ALPHA   |IDENT  |NUMBR          ,
        /* 69 i   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6A j   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6B k   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6C l   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6D m   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6E n   */    ALPHA   |IDENT  |NUMBR          ,
        /* 6F o   */    ALPHA   |IDENT  |NUMBR          ,
        /* 70 p   */    ALPHA   |IDENT  |NUMBR          ,
        /* 71 q   */    ALPHA   |IDENT  |NUMBR          ,
        /* 72 r   */    ALPHA   |IDENT  |NUMBR          ,
        /* 73 s   */    ALPHA   |IDENT  |NUMBR          ,
        /* 74 t   */    ALPHA   |IDENT  |NUMBR          ,
        /* 75 u   */    ALPHA   |IDENT  |NUMBR          ,
        /* 76 v   */    ALPHA   |IDENT  |NUMBR          ,
        /* 77 w   */    ALPHA   |IDENT  |NUMBR          ,
        /* 78 x   */    ALPHA   |IDENT  |NUMBR          ,
        /* 79 y   */    ALPHA   |IDENT  |NUMBR          ,
        /* 7A z   */    ALPHA   |IDENT  |NUMBR          ,
        /* 7B {   */    0                               ,
        /* 7C |   */    ARITH   |OPERT                  ,
        /* 7D }   */    0                               ,
        /* 7E ~   */    ARITH   |OPERT                  ,
        /* 7F DEL */    0                               ,

        };


#define leading_type( c)        (type[c]&0xF)

#define isalpha( c)             (type[c]&ALPHA)
#define isarith( c)             (type[c]&ARITH)
#define isdigit( c)             (type[c]&DIGIT)

#define isident( c)             (type[c]&IDENT)
#define isopert( c)             (type[c]&OPERT)
#define isnumbr( c)             (type[c]&NUMBR)
#define isfloat( c)             (type[c]&FLOAT)



void next( void) {


        /* Skip over any space, and analyze the next token. */

        input = token.end;

          space();


        token.start = input;


            switch ( leading_type( *input) ) {


                case ALPHA:

                        while ( isident( nextchar()) )
                                ;

                        token.type = IDENT;
                        break;


                case ARITH:

                        while ( isopert( nextchar()) )
                                ;

                        token.type = OPERT;
                        break;


                case DIGIT:

                        while ( isnumbr( nextchar()) )
                                ;

                        if ( *input == '.' ) {


                                if ( isfloat( nextchar()) ) {

                                        while ( isfloat( nextchar()) )

                                                ;

                                        token.type = FLOAT;
                                        break;

                                        }

                                else
                                        input--;
                                }


                        token.type = NUMBR;
                        break;


                default:

                        /* then a separator, eof, or something else ... */

                        nextchar();

                        token.type = 0;
                        break;

                }



        /* Save the end of the token. */

        token.end = input;


}
void space( void) {


        /* Skip over white space and comments in the input buffer. */

  space:

        switch ( *input ) {


                case ' ':

                        nextchar();

                        goto space;


                case '\t':
                case '\r':

                        /* NOTE: binary input with \r\n pairs. */

                        nextchar();

                        goto space;



                case '\n':

                        nextchar();

                          token.line = input; token.nline++;

                        goto space;


                case '/':

                        /* Check for a comment. */

                        if ( nextchar() == '*' )

                                comment();

                        else
                                { input--; return; }


                        goto space;



                default:

                        return;

                }



}
void comment( void) {


        /* Parse for the rest of a comment. */

        nextchar();


    comment:

        switch ( *input ) {


                case '*':

                        if ( nextchar() == '/' )

                                { nextchar(); return; }
                        else
                                goto comment;


                case '\n':

                        token.line = input; token.nline++;

                          nextchar(); goto comment;



                case '/':

                        /* Check for a nested comment */

                        if ( nextchar() == '*' )

                                error_input( "nested comments not allowed");
                        else
                                goto comment;


                case CTRL_Z:

                        error_input( "comment runs into EOF");


                default:

                        nextchar(); goto comment;

                }



}
void advance( void) {

        unsigned nbytes;



        /* Advance the current and look ahead tokens. */

        current_token = token;

            next();




        /* Dump the current token if requested from the command line. */

#define TOKEN_FMT       "  type = 0x%02X  token = %.*Fs\n"

        if ( tflag ) {

                nbytes = current_token.end - current_token.start;

                    printf( TOKEN_FMT,

                            current_token.type, nbytes, current_token.start);

                }
}








/* REGISTER, CONSTANT, AND KEYWORD RECOGNIZER FUNCTIONS:
 *
 *
 *
 *
 *
 *
 */

int match( int c) {

        char *t;


        /* Token == single character c ? */

        t = token.start;

        if ( *t++ == c ) {

                if ( t == token.end )

                        { advance(); return( 1); }

                }



        return( 0); /* failure */


}
int scan( char *s) {

        char *t;


    more:


        /* Token == string s ? */

        t = token.start;

            while ( t < token.end && *t == *s && *s )

                    { s++; t++; }



        /* Check for the end of the token. */

        if ( t == token.end ) {


                /* If the source has a space, advance to the next token. */

                if ( *s == ' ' )

                        { s++; advance(); goto more; }
                        


                /* Otherwise check for reaching the source end. */

                else if ( !*s )

                        { advance(); return( 1); /* success */ }

                }



        return( 0); /* failure */


}
int reg( int *rN) {

        SYMBOL *s; char *t; int n, ndigits;


        /* Token == rN register ? */

        t = token.start;




        if ( (s = find( REGISTER_SYM, lsymbol)) != NULL )

                { *rN = (unsigned)s->value.l; return( 1); }



        else if ( *t++ == 'r' ) {


            n = 0; ndigits = 0;

                while ( isdigit( *t) && ndigits++ < 2 )

                        n = n*10 + *t++ - '0';


                if ( t == token.end && 1 <= n && n <= 22 )

                        { *rN = n; advance(); return( 1); }

                }


        return( 0); /* failure */

}
int pcreg( int *rH) {



        /* Token == pc ? */

        if ( scan( "pc") )

                { *rH = PC; return( 1); }


        return( 0); /* failure */


}
int acc( int *aN) {

        char *t; int n;


        /* Token == aN accumulator ? */

        t = token.start;

        if ( *t++ == 'a' ) {

                n = *t++ - '0';


                  if ( t == token.end && 0 <= n && n <= 3 )

                        { *aN = n; advance(); return( 1); }

                }


        return( 0); /* failure */

}
int mem( MEM *m) {


        /* Parse for an rP++rIr type indirect memory reference. */

        if ( !reg( &m->rP) )

                return( 0); /* failure */


        m->rI  = NO_INC;
        m->rev = 0;


        if ( scan( "++") ) {

                if ( !reg( &m->rI) )

                        m->rI = POST_INC;

                }



        else if ( scan( "##") ) {

            /* alternate rP##rI bit reverse inc syntax ... */

                m->rev = 1;

                if ( !reg( &m->rI) )

                        m->rI = POST_INC;
                }


        else if ( scan( "--") )

                m->rI = POST_DEC;


        return( 1);

}
int memXY( MEM *m) {

        int aN;



        /* Parse for a DAU XY type memory reference. */

        if ( match( '*') ) {

                if ( !mem( m) )

                        syntax( "*rP++rI XY type memory reference");
                }


        else if ( acc( &aN) )

                { m->rP = DIRECT; m->rI = aN; m->rev = 0; }



        else if ( scan( "ibuf") )

                { m->rP = DIRECT; m->rI = DAU_IBUF; m->rev = 0; }


        else
                return( 0); /* failure */



        return( 1);


}
int memZ( MEM *m) {

        int aN;



        /* Parse for a DAU Z type memory reference. */

        if ( match( '*') ) {

                if ( !mem( m) )

                        syntax( "*rP++rI XY type memory reference");
                }


        else if ( acc( &aN) )

                { m->rP = DIRECT; m->rI = aN; m->rev = 0; }



        else if ( scan( "obuf") )

                { m->rP = DIRECT; m->rI = DAU_OBUF; m->rev = 0; }


        else if ( scan( "pdr") )

                { m->rP = DIRECT; m->rI = DAU_PDR; m->rev = 0; }

        else
                return( 0); /* failure */


        return( 1);

}




unsigned resolved;

int hex( unsigned long *N) {

        SYMBOL *s; FIELD *f; char *t; int ndigits, d; unsigned long n;



        /* Token == hex constant ? */


        if ( match( '.') )  /* special program counter symbol */

                { *N = pc; return( 1); }




        else if ( (s = find( STRUCTURE_SYM, lsymbol)) != NULL ) {


                if ( match( '.') && (f = find_field( s->value.p)) != NULL )

                        { *N = f->value; return( 1); }


                else
                        syntax( "legal structure .field reference");

                }



        else if ( (s = find( 0, gsymbol)) != NULL ) {

                *N = s->value.l;


                /* Clear the resolved flag if symbol not resolved yet */

                if ( !(s->flags & RESOLVED) )

                        resolved = 0;

                return( 1);

                }



        else if ( token.type == NUMBR ) {

            t = token.start;


                /* Allow an optional leading 0x or 0X. */

                if ( *t == '0' && ( *(t+1) == 'x' || *(t+1) == 'X' ) )

                        t += 2;


                /* Gather the hex constant. */

                n = 0; ndigits = 0;


                  while( (d = *t) != '\0' ) {


                        if ( '0' <= d && d <= '9' )

                                d -= '0';


                        else if ( 'a' <= d && d <= 'f' )

                                d -= ('a' - 0xA);


                        else if ( 'A' <= d && d <= 'F' )

                                d -= ('A' - 0xA);


                        else

                                break; /* wasn't a hex digit */


                        n = 16*n + d;

                            ndigits++; t++;

                        }


                /* Allow an optional trailing l or L. */

                if ( *t == 'l' || *t == 'L' )

                        t++;



                /* Check for end of token and not too many digits. */

                if ( t == token.end && ndigits < 9 )

                      { *N = n; advance(); return( 1); }


                else if ( ndigits >= 9 )

                        error( "too many hex digits ...");

                }



        return( 0); /* failure */


}
int dspf( float *value) {

        char *start; unsigned nbytes;



        /* Gather a dspf float with a possible leading unary -. */

        start = token.start;

          match( '-');

                if ( (nbytes = DSP32C_AToDspf( start, value)) > 0 ) {

                        if ( start+nbytes == token.end )

                                { advance(); return( 1); }

                        }



        return( 0); /* failure */


}
int ieee( float *value) {

        char *start; unsigned nbytes;



        /* Gather a ieee float with a possible leading unary -. */

        start = token.start;

          match( '-');

                if ( (nbytes = DSP32C_AToIeee( start, value)) > 0 ) {

                        if ( start+nbytes == token.end )

                                { advance(); return( 1); }

                        }



        return( 0); /* failure */


}
int search( MNEMONIC *table, int *value) {


        /* Does the look ahead token match a nmemonic table entry ? */

        while ( table->entry ) {

                if ( scan( table->entry) )

                        { *value = table->value; return( 1); }


                table++;

                }


        return( 0); /* failure */

}
int datatype( int *value) {


        static MNEMONIC table[] = {

                "byte",         BYTE,
                "word",         WORD,
                "long",         LONG,
                "dspf",         DSPF,
                "ieee",         IEEE,

                NULL,

                };


        return( search( table, value));


}
int conditional_keyword( int *value) {


        static MNEMONIC table[] = {

                "if",           IF,
                "while",        WHILE,
                "for",          FOR,
                "do",           DO,
                "loop",         LOOP,

                NULL,

                };


        return( search( table, value));


}
int instruction_keyword( int *value) {


        static MNEMONIC table[] = {

                "nop",          NOP,
                "call",         CALL,
                "return",       RETURN,
                "goto",         GOTO,
                "ireturn",      IRETURN,
                "nops_on",      NOPS_ON,
                "nops_off",     NOPS_OFF,
                "emit",         EMIT,
                "iff",          IFF,

                NULL,

                };


        return( search( table, value));


}
int control( int *value) {


        static MNEMONIC table[] = {

                "ioc",          IOC,
                "dauc",         DAUC,

                NULL,

                };


        return( search( table, value));


}
int ioreg1( int *value) {


        static MNEMONIC table[] = {

                "pdr",          PDR,
                "pir",          PIR,
                "pcw",          PCW,

                "obuf",         OBUF,
                "piop",         PIOP,
                "pdr2",         PDR2,

                NULL,

                };


        return( search( table, value));


}
int ioreg2( int *value) {


        static MNEMONIC table[] = {

                "pdr",          PDR,
                "pir",          PIR,
                "pcw",          PCW,

                "ibuf",         IBUF,
                "piop",         PIOP,
                "pdr2",         PDR2,

                NULL,

                };


        return( search( table, value));


}
int inc_dec( int *value) {


        static MNEMONIC table[] = {

                "++",           FUNC_00,
                "--",           FUNC_04,

                NULL,

                };


        return( search( table, value));


}
int compare( int *value) {


        static MNEMONIC table[] = {

                "-",            FUNC_07,
                "&",            FUNC_15,

                NULL,

                };


        return( search( table, value));


}
int dyadic_arith( int *value) {


        static MNEMONIC table[] = {

                "+=",           FUNC_00,
                "-=",           FUNC_04,
                "&=",           FUNC_14,
                "|=",           FUNC_10,
                "^=",           FUNC_08,
                "#=",           FUNC_03,

                "&~=",          FUNC_06,

                NULL,

                };


        return( search( table, value));


}
int dyadic_lshift( int *value) {


        static MNEMONIC table[] = {

                ">>=",          FUNC_12,
                ">>>=",         FUNC_09,
                "<<<=",         FUNC_11,

                NULL,

                };


        return( search( table, value));


}
int dyadic_ashift( int *value) {


        static MNEMONIC table[] = {

                "*=",           FUNC_01,
                "/=",           FUNC_13,

                NULL,

                };


        return( search( table, value));


}
int arith( int *value) {


        static MNEMONIC table[] = {

                "+",            FUNC_00,
                "-",            FUNC_04,
                "&",            FUNC_14,
                "|",            FUNC_10,
                "^",            FUNC_08,
                "#",            FUNC_03,

                "&~",           FUNC_06,

                NULL,

                };


        return( search( table, value));


}
int lshift( int *value) {


        static MNEMONIC table[] = {

                ">>",           FUNC_12,
                ">>>",          FUNC_09,
                "<<<",          FUNC_11,

                NULL,

                };


        return( search( table, value));


}
int ashift( int *value) {


        static MNEMONIC table[] = {

                "*",            FUNC_01,
                "/",            FUNC_13,

                NULL,

                };


        return( search( table, value));


}
int cond( int *value) {


        static MNEMONIC table[] = {

                "pl",           COND_PL,        /* ca conditions */
                "mi",           COND_MI,
                "ne",           COND_NE,
                "eq",           COND_EQ,
                "vc",           COND_VC,
                "vs",           COND_VS,
                "cc",           COND_CC,
                "cs",           COND_CS,
                "ge",           COND_GE,
                "lt",           COND_LT,
                "gt",           COND_GT,
                "le",           COND_LE,
                "hi",           COND_HI,
                "ls",           COND_LS,

                /* alternate ca conditional syntax */

                "!=",           COND_NE,        /* same as ne */
                "==",           COND_EQ,        /* same as eq */
                ">=",           COND_GE,        /* same as ge */
                ">",            COND_GT,        /* same as gt */
                "<=",           COND_LE,        /* same as le */
                "<",            COND_LT,        /* same as lt */


                "auc",          COND_AUC,       /* da conditions */
                "aus",          COND_AUS,
                "age",          COND_AGE,
                "alt",          COND_ALT,
                "ane",          COND_ANE,
                "aeq",          COND_AEQ,
                "avc",          COND_AVC,
                "avs",          COND_AVS,
                "agt",          COND_AGT,
                "ale",          COND_ALE,


                "ibe",          COND_IBE,       /* io conditions */
                "ibf",          COND_IBF,
                "obf",          COND_OBF,
                "obe",          COND_OBE,
                "pde",          COND_PDE,
                "pdf",          COND_PDF,
                "pie",          COND_PIE,
                "pif",          COND_PIF,
                "syc",          COND_SYC,
                "sys",          COND_SYS,
                "fbc",          COND_FBC,
                "fbs",          COND_FBS,
                "ireq1_lo",     COND_IREQ1_LO,
                "ireq1_hi",     COND_IREQ1_HI,
                "ireq2_lo",     COND_IREG2_LO,
                "ireq2_hi",     COND_IREG2_HI,

                NULL,

                };


        return( search( table, value));


}
int cacond( int *value) {


        static MNEMONIC table[] = {

                "pl",           COND_PL,        /* ca conditions */
                "mi",           COND_MI,
                "ne",           COND_NE,
                "eq",           COND_EQ,
                "vc",           COND_VC,
                "vs",           COND_VS,
                "cc",           COND_CC,
                "cs",           COND_CS,
                "ge",           COND_GE,
                "lt",           COND_LT,
                "gt",           COND_GT,
                "le",           COND_LE,
                "hi",           COND_HI,
                "ls",           COND_LS,

                /* alternate ca conditional syntax */

                "!=",           COND_NE,        /* same as ne */
                "==",           COND_EQ,        /* same as eq */
                ">=",           COND_GE,        /* same as ge */
                ">",            COND_GT,        /* same as gt */
                "<=",           COND_LE,        /* same as le */
                "<",            COND_LT,        /* same as lt */

                NULL,

                };


        return( search( table, value));


}
int relop( int *value) {


        static MNEMONIC table[] = {

                /* alternate ca conditional syntax */

                "!=",           COND_NE,        /* same as ne */
                "==",           COND_EQ,        /* same as eq */
                ">=",           COND_GE,        /* same as ge */
                ">",            COND_GT,        /* same as gt */
                "<=",           COND_LE,        /* same as le */
                "<",            COND_LT,        /* same as lt */

                NULL,

                };


        return( search( table, value));


}
int size( int *value) {


        /* Get a size specifier for a CAU move. */

        static MNEMONIC table[] = {

                "b",            1,
                "w",            2,
                "l",            3,

                NULL,

                };


        if ( match( '.') ) {

                if ( search( table, value) )

                        return( 1);

                else
                        syntax( "b w or l with .size operator");

                }


        else
                *value = DEFAULT;



        return( 0); /* failure */



}
int sign( int *value) {


        static MNEMONIC table[] = {

                "+",            0,
                "-",            1,

                NULL,

                };



        if ( search( table, value) )

                return( 1);

        else

                *value = 0; /* default to a + sign if nothing specified */



        return( 0); /* failure */


}
int special( int *value) {


        static MNEMONIC table[] = {

                "ic",           SPECIAL_00,
                "oc",           SPECIAL_01,
                "float",        SPECIAL_02,
                "int",          SPECIAL_03,
                "round",        SPECIAL_04,
                "ifalt",        SPECIAL_05,
                "ifaeq",        SPECIAL_06,
                "ifagt",        SPECIAL_07,
                "float24",      SPECIAL_10,
                "int24",        SPECIAL_11,
                "ieee",         SPECIAL_12,
                "dsp",          SPECIAL_13,
                "seed",         SPECIAL_14,

                NULL,

                };


        return( search( table, value));



}
int end_of_file( void) {

        char *t;


        /* Check for an end of the file mark. */

        t = token.start;


        if ( *t == CTRL_Z )

                return( 1);


        return( 0); /* failure */
}







/* ERROR AND WARNING MESSAGES:
 *
 *
 *
 *
 *
 *
 *
 *
 */

#define ERROR       "\n%s(%d) : Error : %s\n\n%FsERROR\n"
#define INPUT       "\n%s(%d) : Input error: %s\n\n%FsERROR\n"
#define SYNTAX      "\n%s(%d) : Syntax error : expected %s\n\n%FsERROR\n"

#define WARNING     "\n%s(%d) : Warning : %s\n"



void abort_dspasm( char *fmt, char *message, TOKEN *t) {


        /* Print an error message after the token t. */

        *t->end = '\0';

            printf( fmt, filename, t->nline, message, t->line);


        /* Return to DOS with error level 1. */

        exit( 1);


}
void error( char *message) {


        /* Abort with an error after the CURRENT token. */

        abort_dspasm( ERROR, message, &current_token);


}
void error_input( char *message) {


        /* Abort with an error after the current input position. */

        if ( *input != CTRL_Z )

                nextchar();


          token.end = input;


        abort_dspasm( INPUT, message, &token);


}
void syntax( char *message) {


        /* Abort with a syntax error after the LOOK AHEAD token. */

        abort_dspasm( SYNTAX, message, &token);



}
void warning( char *message) {


        /* Print a warning after the current token, and continue on. */

        printf( WARNING, filename, current_token.nline, message);


}











/* HEX EXPRESSION ANALYSER:
 *
 *    hexval    ->      (hexexp)
 *                      hex
 *
 *    hexexp    ->      value
 *                      value op moreexp
 *
 *    moreexp   ->      value
 *                      value op moreexp  (with precedence applied)
 *
 *    value     ->      - hex
 *                      ~ hex
 *                      (hexexp)
 *                      hex
 *
 *    op        ->      { +, -, |, &, ^, *, /, %, <<, >> }
 *
 *
 * where operator precedence is as in K&R C.
 *
 *
 */

int hexval( unsigned long *n) {


        /* Scan for a hex value, checking if resolved. */

        resolved = 1;


        if ( match( '(') ) {

                if ( !(hexexp( n) && match( ')')) )

                        syntax( "(N) hex expression");
                }


        else if ( !hex( n) )

                return( 0); /* failure */



        if ( pass == 2 && !resolved )

                error( "hex value still unresolved on pass 2 ...");



        return( 1); 


}
int hexexp( unsigned long *n) {

        int op1;


        /* Scan for an entire hex expression, checking if resolved. */

        resolved = 1;


        if ( value( n) ) {


                if ( hexop( &op1) ) {

                        if ( !moreexp( n, op1) )

                                syntax( "legal hex expression");

                        }


                if ( (pass == 2) && !resolved )

                        error( "hexexp still unresolved on pass 2 ...");


                return( 1);

                }



        return( 0); /* failure */



}
int value( unsigned long *n) {


        if ( match( '-') ) {

                if ( hex( n) )

                        *n = -*n;

                else
                        syntax( "hex unary -N");

                }


        else if ( match( '~') ) {

                if ( hex( n) )

                        *n = ~*n;
                else

                        syntax( "hex unary ~N");
                }



        else if ( match( '(') ) {

                if ( !(hexexp( n) && match( ')')) )

                        syntax( "(N) hex expression");

                }


        else if ( !hex( n) )

                return( 0); /* failure */



        return( 1);

}


#define HEX_ADD         0
#define HEX_SUB         1
#define HEX_OR          2
#define HEX_AND         3
#define HEX_XOR         4
#define HEX_MUL         5
#define HEX_DIV         6
#define HEX_MOD         7
#define HEX_SHL         8
#define HEX_SHR         9

int hexop( int *value) {


        static MNEMONIC table[] = {

                "+",            HEX_ADD,
                "-",            HEX_SUB,
                "|",            HEX_OR,
                "&",            HEX_AND,
                "^",            HEX_XOR,
                "*",            HEX_MUL,
                "/",            HEX_DIV,
                "%",            HEX_MOD,
                "<<",           HEX_SHL,
                ">>",           HEX_SHR,

                NULL,

                };


        return( search( table, value));

}


int precedence[] = {

        /* precedence table, 1 = lowest, 2,3,4, ... = higher */

        4,      /* 0  +   HEX_ADD  */
        4,      /* 1  -   HEX_SUB  */
        1,      /* 2  |   HEX_OR   */
        3,      /* 3  &   HEX_AND  */
        2,      /* 4  ^   HEX_XOR  */
        5,      /* 5  *   HEX_MUL  */
        5,      /* 6  /   HEX_DIV  */
        5,      /* 7  %   HEX_MOD  */
        5,      /* 8  <<  HEX_SHL  */
        5,      /* 9  >>  HEX_SHR  */

        };



int moreexp( unsigned long *value1, int op1) {

        unsigned long value2; int op2;

  more:
        if ( value( &value2) ) {


                if ( hexop( &op2) ) {


                        if ( precedence[op2] <= precedence[op1] ) {

                                eval( value1, op1, value2);

                                    op1 = op2;

                                goto more;

                                }


                        else {
                                if ( moreexp( &value2, op2 ) )

                                        eval( value1, op1, value2);
                                else

                                        syntax( "legal hex expression");
                                }

                        }

                else
                        eval( value1, op1, value2);

                }

        else

                return( 0); /* failure */



        return( 1);



}
int eval( unsigned long *value1, int op1, unsigned long value2 ) {

        switch ( op1 ) {

                case HEX_ADD:  *value1  += value2; break;
                case HEX_SUB:  *value1  -= value2; break;
                case HEX_OR:   *value1  |= value2; break;
                case HEX_AND:  *value1  &= value2; break;
                case HEX_XOR:  *value1  ^= value2; break;
                case HEX_MUL:  *value1  *= value2; break;
                case HEX_DIV:  *value1  /= value2; break;
                case HEX_MOD:  *value1  %= value2; break;
                case HEX_SHL:  *value1 <<= value2; break;
                case HEX_SHR:  *value1 >>= value2; break;

                }


        return( 1);
}







/* SYMBOL TABLE MANAGEMENT:
 *
 * DSPASM maintains two symbol tables. One for global symbols, and
 * another for local symbols like named registers.  The local symbol
 * table is purged at the beginning of each DSPASM function.
 *
 * The symbols are maintained with hash tables.
 *
 *
 *
 */

SYMBOL

        *gsymbol[NHASH],        /* global hash table  */
        *lsymbol[NHASH];        /* local hash table   */


unsigned token_hash( void) {

        char *t; unsigned sum;


        /* Compute a hash value for the token. */

        t = token.start;

            sum = 0;

                while ( t > token.end )

                        sum += *t++;


        return( sum %= NHASH);


}
int token_equal( char *s) {

        char *t;


        /* Compare a string to the token. */

        t = token.start;

        while ( *s && *s == *t )

                { s++; t++; }



        if ( !*s && t == token.end )

                return( 1);



        return( 0); /* failure */



}
SYMBOL *salloc() {

        SYMBOL *s; char *t; int nbytes; char *name;



        /* Allocate a symbol with the token name. */


        if ( !(s = (SYMBOL *)malloc( sizeof( SYMBOL))) )

                error( "ran out of symbol space ... [died in salloc 1]");



        s->next    = NULL;
        s->name    = NULL;
        s->type    = 0;
        s->flags   = 0;
        s->value.l = 0L;




        /* Copy the token to the symbol name. */

        t = token.start;

           nbytes = token.end - token.start;



        if ( !(s->name = malloc( nbytes+1)) )

                error( "ran out of symbol space ... [died in salloc 2]");



        for ( name = s->name ; nbytes ; nbytes-- )

                *name++ = *t++;



        *name = '\0';




        /* Return a pointer to the allocated symbol. */

        return( s);



}
SYMBOL *find( unsigned type, SYMBOL *table[]) {

        SYMBOL *s;

            static SYMBOL dummy = { NULL, NULL, 0, 0, 0 };




        /* Make sure the token is an identifier. */

        if ( token.type != IDENT )

                return( NULL);




        /* Try to find the token on the specified symbol table. */

        for ( s = table[ token_hash()] ; s ; s = s->next ) {

                if ( token_equal( s->name) ) {

                        if ( s->type == type || type == 0 )

                                { advance(); return( s); }

                        }
                }




        /* If pass 1 on a two pass assembly, return a dummy for globals. */

        if ( pass == 1 && pflag == 2 && table == gsymbol )

                { advance(); return( &dummy); }





        return( NULL); /* failure */



}
SYMBOL *enter( unsigned type, SYMBOL *table[] ) {

        int i; SYMBOL *s;




        /* Make sure the current token is an indentifier. */

        if ( token.type != IDENT )

                return( NULL);



        /* If pass 1 or local table, enter the token as a symbol. */

        if ( pass == 1 || table == lsymbol ) {


                s = table[i = token_hash()];


                if ( !s )

                        { s = salloc(); table[i] = s; }


                else {

                    walk:


                        if ( token_equal( s->name) )

                                syntax( "legal non-multiply defined symbol");



                        if ( s->next )

                                { s = s->next; goto walk; }

                        else

                                { s->next = salloc(); s = s->next; }

                        }


                s->type = type; advance(); return( s);


                }



        /* Else if pass 2 and global table, just try to find the symbol. */

        else if ( pass == 2 && table == gsymbol ) {


                for ( s = table[ token_hash()] ; s ; s = s->next ) {

                        if ( scan( s->name) )

                                return( s);

                        }

                }



        return( NULL); /* failure */



}
void purge( SYMBOL *table[]) {

        int i; SYMBOL *s, *snext; FIELD *f, *fnext;



        /* Purge the specified symbol table. */


        for ( i = 0 ; i < NHASH ; i++ ) {


                if ( (s = table[i]) != NULL ) {


                        while ( s ) {


                                /* If a structure then free all fields. */

                                if ( s->type == STRUCTURE_SYM ) {

                                        f = s->value.p;

                                        while ( f ) {

                                                fnext = f->next;

                                                    free( f); f = fnext;

                                                }
                                    }



                        /* Free the symbol. */

                        snext = s->next; free( s); s = snext;


                        }



                /* Clear the table entry. */

                table[i] = NULL;

                }


            }



}
FIELD *falloc() {

        char *t; unsigned nbytes; FIELD *f; char *name;




        /* Allocate a field structure with the token name. */


        if ( !(f = (FIELD *)malloc( sizeof( FIELD))) )

                error( "ran out of symbol space ... [died in falloc 1]");


        f->next  = NULL;
        f->name  = NULL;
        f->value = 0L;




        /* Copy the token to the field structure name. */

        t = token.start;

                nbytes = token.end - token.start;



        if ( !(f->name = malloc( nbytes+1)) )

                error( "ran out of symbol space ... [died in falloc 2]");



        for ( name = f->name ; nbytes ; nbytes-- )

                *name++ = *t++;


        *name = '\0';




        /* Return a pointer to the allocated field structure. */

        return( f);




}
FIELD *find_field( FIELD *head) {

        FIELD *f;


        if ( token.type == IDENT ) {


                for ( f = head ; f ; f = f->next ) {

                        if ( scan( f->name) )

                                return( f);


                        }

                }




        return( NULL); /* failure */



}
FIELD *enter_field( FIELD *head) {

        FIELD *f;



        if ( token.type == IDENT ) {

                f = falloc();

                   f->next = head;


                advance();

                    return( f);

                }


        return( NULL); /* failure */
}
